@@ -11,6 +11,7 @@
  *  Author: Guillaume Roguez <guillaume.roguez@savoirfairelinux.com>
  *  Author: Adrien Béraud <adrien.beraud@savoirfairelinux.com>
  *  Author: Philippe Gorley <philippe.gorley@savoirfairelinux.com>
+ *  Author: Aline Gondim Santos <aline.gondimsantos@savoirfairelinux.com>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -64,6 +65,10 @@ using random_device = dht::crypto::random_device;
 #include "audio/sound/dtmf.h"
 #include "audio/ringbufferpool.h"
 
+#ifdef ENABLE_PLUGIN
+#include "plugin/jamipluginmanager.h"
+#endif
+
 #ifdef ENABLE_VIDEO
 #include "client/videomanager.h"
 #include "video/video_scaler.h"
@@ -288,7 +293,7 @@ struct Manager::ManagerPimpl
      * Play one tone
      * @return false if the driver is uninitialize
      */
-    void playATone(Tone::TONEID toneId);
+    void playATone(Tone::ToneId toneId);
 
     int getCurrentDeviceIndex(DeviceType type);
 
@@ -415,6 +420,10 @@ struct Manager::ManagerPimpl
 #endif
 
     std::unique_ptr<SIPVoIPLink> sipLink_;
+#ifdef ENABLE_PLUGIN
+    /* Jami Plugin Manager */
+    JamiPluginManager jami_plugin_manager;
+#endif
 };
 
 Manager::ManagerPimpl::ManagerPimpl(Manager& base)
@@ -467,7 +476,7 @@ Manager::ManagerPimpl::parseConfiguration()
  * Multi Thread
  */
 void
-Manager::ManagerPimpl::playATone(Tone::TONEID toneId)
+Manager::ManagerPimpl::playATone(Tone::ToneId toneId)
 {
     if (not base_.voipPreferences.getPlayTones())
         return;
@@ -689,6 +698,9 @@ Manager::Manager()
     , shortcutPreferences()
 #ifdef ENABLE_VIDEO
     , videoPreferences()
+#endif
+#ifdef ENABLE_PLUGIN
+    , pluginPreferences()
 #endif
     , callFactory()
     , accountFactory()
@@ -1692,6 +1704,9 @@ Manager::saveConfig()
         audioPreference.serialize(out);
 #ifdef ENABLE_VIDEO
         videoPreferences.serialize(out);
+#endif
+#ifdef ENABLE_PLUGIN
+        pluginPreferences.serialize(out);
 #endif
         shortcutPreferences.serialize(out);
 
@@ -2021,7 +2036,7 @@ Manager::stopTone()
 void
 Manager::playTone()
 {
-    pimpl_->playATone(Tone::TONE_DIALTONE);
+    pimpl_->playATone(Tone::ToneId::DIALTONE);
 }
 
 /**
@@ -2030,7 +2045,7 @@ Manager::playTone()
 void
 Manager::playToneWithMessage()
 {
-    pimpl_->playATone(Tone::TONE_CONGESTION);
+    pimpl_->playATone(Tone::ToneId::CONGESTION);
 }
 
 /**
@@ -2039,7 +2054,7 @@ Manager::playToneWithMessage()
 void
 Manager::congestion()
 {
-    pimpl_->playATone(Tone::TONE_CONGESTION);
+    pimpl_->playATone(Tone::ToneId::CONGESTION);
 }
 
 /**
@@ -2048,7 +2063,7 @@ Manager::congestion()
 void
 Manager::ringback()
 {
-    pimpl_->playATone(Tone::TONE_RINGTONE);
+    pimpl_->playATone(Tone::ToneId::RINGTONE);
 }
 
 /**
@@ -2746,6 +2761,15 @@ Manager::loadAccountMap(const YAML::Node& node)
         shortcutPreferences.unserialize(node);
 #ifdef ENABLE_VIDEO
         videoPreferences.unserialize(node);
+#endif
+#ifdef ENABLE_PLUGIN
+        pluginPreferences.unserialize(node);
+
+        std::vector<std::string> loadedPlugins = pluginPreferences.getLoadedPlugins();
+        for (const std::string& plugin : loadedPlugins)
+        {
+            jami::Manager::instance().getJamiPluginManager().loadPlugin(plugin);
+        }
 #endif
     } catch (const YAML::Exception &e) {
         JAMI_ERR("%s: Preferences node unserialize error: ", e.what());
@@ -2933,13 +2957,33 @@ Manager::sendRegister(const std::string& accountID, bool enable)
         acc->doUnregister();
 }
 
+bool
+Manager::isPasswordValid(const std::string& accountID, const std::string& password)
+{
+    const auto acc = getAccount<JamiAccount>(accountID);
+    if (!acc)
+        return false;
+    return acc->isPasswordValid(password);
+}
+
 uint64_t
 Manager::sendTextMessage(const std::string& accountID, const std::string& to,
                          const std::map<std::string, std::string>& payloads)
 {
     if (const auto acc = getAccount(accountID)) {
         try {
+#ifdef ENABLE_PLUGIN
+            auto& convManager = jami::Manager::instance().getJamiPluginManager()
+                    .getConversationServicesManager();
+            std::shared_ptr<jami::ConversationMessage> cm =
+                    std::make_shared<jami::ConversationMessage>(accountID, to,
+                                                                const_cast<std::map<std::string,
+                                                                std::string>&>(payloads));
+            convManager.sendTextMessage(cm);
+            return acc->sendTextMessage(cm->to_, cm->data_);
+#else
             return acc->sendTextMessage(to, payloads);
+#endif //ENABLE_PLUGIN
         } catch (const std::exception& e) {
             JAMI_ERR("Exception during text message sending: %s", e.what());
         }
@@ -3089,6 +3133,13 @@ Manager::sipVoIPLink() const
     return *pimpl_->sipLink_;
 }
 
+#ifdef ENABLE_PLUGIN
+JamiPluginManager&
+Manager::getJamiPluginManager() const
+{
+    return pimpl_->jami_plugin_manager;
+}
+#endif
 
 std::map<std::string, std::string>
 Manager::getNearbyPeers(const std::string& accountID)
